@@ -1,25 +1,34 @@
+from __future__ import print_function
+
 from phply import phplex
-from phply.phpparse import parser
+from phply.phpparse import make_parser
 from phply.phpast import *
 
 import nose.tools
 import pprint
 
-def eq_ast(input, expected, filename=None):
+parser = make_parser()
+
+def eq_ast(input, expected, filename=None, with_top_lineno=False):
     lexer = phplex.lexer.clone()
     lexer.filename = filename
     output = parser.parse(input, lexer=lexer)
     resolve_magic_constants(output)
 
-    print 'Parser output:'
+    print('Parser output:')
     pprint.pprint(output)
-    print
+    print()
 
-    print 'Node by node:'
+    print('Node by node:')
     for out, exp in zip(output, expected):
-        print '\tgot:', out, '\texpected:', exp
+        print('\tgot:', out, '\texpected:', exp)
         nose.tools.eq_(out, exp)
 
+        # compare line numbers, but only for top elements
+        if with_top_lineno:
+            print('\tgot line:', out.lineno, '\texpected:', exp.lineno)
+            nose.tools.eq_(out.lineno, exp.lineno)
+
     assert len(output) == len(expected), \
            'output length was %d, expected %s' % (len(output), len(expected))
 
@@ -45,8 +54,8 @@ def test_open_tag_with_echo():
 def test_exit():
     input = '<?php exit; exit(); exit(123); die; die(); die(456); ?>'
     expected = [
-        Exit(None), Exit(None), Exit(123),
-        Exit(None), Exit(None), Exit(456),
+        Exit(None, 'exit'), Exit(None, 'exit'), Exit(123, 'exit'),
+        Exit(None, 'die'), Exit(None, 'die'), Exit(456, 'die'),
     ]
     eq_ast(input, expected)
 
@@ -56,12 +65,14 @@ def test_isset():
         isset($b->c);
         isset($d['e']);
         isset($f, $g);
+        isset($h->m()['i1']['i2']);
     ?>"""
     expected = [
         IsSet([Variable('$a')]),
         IsSet([ObjectProperty(Variable('$b'), 'c')]),
         IsSet([ArrayOffset(Variable('$d'), 'e')]),
         IsSet([Variable('$f'), Variable('$g')]),
+        IsSet([ArrayOffset(ArrayOffset(MethodCall(Variable('$h'), 'm', []), 'i1'), 'i2')]),
     ]
     eq_ast(input, expected)
 
@@ -138,6 +149,7 @@ def test_string_unescape():
         '\r\n\t\\\'';
         "\r\n\t\\\"";
     ?>"""
+    # TODO: "\x97\x[0-9]";
     expected = [
         r"\r\n\t\'",
         "\r\n\t\\\"",
@@ -233,7 +245,7 @@ EOT;
                                                                  'variables')),
                                          '.\n'),
                                 'This'),
-                       ' is not the EOT; this is:\n')]),
+                       ' is not the EOT; this is:')]),
     ]
     eq_ast(input, expected)
 
@@ -332,29 +344,45 @@ def test_foreach():
         }
         foreach (complex($expression) as &$ref)
             $ref++;
-        foreach ($what as &$de => &$dealy):
+        foreach ($what as $de => &$dealy):
             yo();
             yo();
         endforeach;
+        foreach ($foo as $bar[0]) {}
     ?>"""
     expected = [
-        Foreach(Variable('$foo'), None, ForeachVariable('$bar', False),
+        Foreach(Variable('$foo'), None, ForeachVariable(Variable('$bar'), False),
                 Block([Echo([Variable('$bar')])])),
         Foreach(Variable('$spam'),
-                ForeachVariable('$ham', False),
-                ForeachVariable('$eggs', False),
+                Variable('$ham'),
+                ForeachVariable(Variable('$eggs'), False),
                 Block([Echo([BinaryOp('.',
                                       BinaryOp('.', Variable('$ham'), ': '),
                                       Variable('$eggs'))])])),
         Foreach(FunctionCall('complex', [Parameter(Variable('$expression'),
                                                    False)]),
-                None, ForeachVariable('$ref', True),
+                None, ForeachVariable(Variable('$ref'), True),
                 PostIncDecOp('++', Variable('$ref'))),
         Foreach(Variable('$what'),
-                ForeachVariable('$de', True),
-                ForeachVariable('$dealy', True),
+                Variable('$de'),
+                ForeachVariable(Variable('$dealy'), True),
                 Block([FunctionCall('yo', []),
                        FunctionCall('yo', [])])),
+        Foreach(Variable('$foo'),
+                None,
+                ForeachVariable(ArrayOffset(Variable('$bar'), 0), False),
+                Block([])),
+    ]
+    eq_ast(input, expected)
+
+def test_foreach_with_lists():
+    input = r"""<?
+        foreach ($foo as list($bar, $baz)) {}
+        foreach ($foo as $k => list($bar, $baz)) {}
+    ?>"""
+    expected = [
+        Foreach(Variable('$foo'), None, ForeachVariable([Variable('$bar'), Variable('$baz')], False), Block([])),
+        Foreach(Variable('$foo'), Variable('$k'), ForeachVariable([Variable('$bar'), Variable('$baz')], False), Block([])),
     ]
     eq_ast(input, expected)
 
@@ -406,7 +434,7 @@ def test_classes():
         class Stub {}
     ?>"""
     expected = [
-        Class('Clown', 'final', 'Unicycle', ['RedNose', 'FacePaint'], [
+        Class('Clown', 'final', 'Unicycle', ['RedNose', 'FacePaint'], [], [
             ClassConstants([ClassConstant('the', 'only'),
                             ClassConstant('constant', 'is')]),
             ClassConstants([ClassConstant('change', 'chump')]),
@@ -421,7 +449,7 @@ def test_classes():
                    [Return(BinaryOp('.', Variable('$arg1'), Variable('$arg2')))],
                    False),
         ]),
-        Class('Stub', None, None, [], []),
+        Class('Stub', None, None, [], [], []),
     ]
     eq_ast(input, expected)
 
@@ -482,7 +510,44 @@ def test_exceptions():
             Catch('Exception', Variable('$e'), [
                 Echo(['Problem?']),
             ]),
-        ])
+        ],
+        None)
+    ]
+    eq_ast(input, expected)
+
+def test_catch_finally():
+    input = r"""<?
+        try {
+            1;
+        } catch (Exception $e) {
+            2;
+        } finally {
+            3;
+        }
+    ?>"""
+    expected = [
+        Try([
+            1
+        ], [
+            Catch('Exception', Variable('$e'), [
+                2
+            ]),
+        ],
+        Finally([3]))
+    ]
+    eq_ast(input, expected)
+
+def test_just_finally():
+    input = r"""<?
+        try {
+        } finally {
+            1;
+        }
+    ?>"""
+    expected = [
+        Try([
+        ], [],
+        Finally([1]))
     ]
     eq_ast(input, expected)
 
@@ -513,11 +578,13 @@ def test_instanceof():
             echo '$foo is a bar';
         }
         $foo instanceof $bar;
+        $foo instanceof static;
     ?>"""
     expected = [
         If(BinaryOp('instanceof', Variable('$foo'), Constant('Bar')),
            Block([Echo(['$foo is a bar'])]), [], None),
         BinaryOp('instanceof', Variable('$foo'), Variable('$bar')),
+        BinaryOp('instanceof', Variable('$foo'), 'static'),
     ]
     eq_ast(input, expected)
 
@@ -554,6 +621,7 @@ def test_casts():
         (iNteger) $x;
         (sTring) $x;
         (uNset) $x;
+        (bInary) $x;
     ?>"""
     expected = [
         Cast('array', Variable('$x')),
@@ -566,6 +634,7 @@ def test_casts():
         Cast('int', Variable('$x')),
         Cast('string', Variable('$x')),
         Cast('unset', Variable('$x')),
+        Cast('binary', Variable('$x')),
     ]
     eq_ast(input, expected)
 
@@ -686,7 +755,7 @@ def test_magic_constants():
                 MagicConstant('__FUNCTION__', 'Shmamespace\\p'), ': '),
                 Variable('$x')), '\n')])
         ], False),
-        Class('Bar', None, None, [],
+        Class('Bar', None, None, [], [],
               [Method('__construct', [], [],
                       [FunctionCall('p', [Parameter(MagicConstant('__LINE__', 10), False)]),
                        FunctionCall('p', [Parameter(MagicConstant('__DIR__', '/my/dir'), False)]),
@@ -700,7 +769,7 @@ def test_magic_constants():
 
 def test_type_hinting():
     input = r"""<?
-    function foo(Foo $var1, Bar $var2=1, Quux &$var3, Corge &$var4=1) {
+    function foo(Foo $var1, Bar $var2=1, Quux &$var3, Corge &$var4=1, array &$var5=array()) {
     }
     ?>""";
     expected = [
@@ -708,8 +777,215 @@ def test_type_hinting():
             [FormalParameter('$var1', None, False, 'Foo'),
              FormalParameter('$var2', 1, False, 'Bar'),
              FormalParameter('$var3', None, True, 'Quux'),
-             FormalParameter('$var4', 1, True, 'Corge')],
+             FormalParameter('$var4', 1, True, 'Corge'),
+             FormalParameter('$var5', Array([]), True, 'array')],
             [],
             False)]
     eq_ast(input, expected)
 
+def test_static_scalar_class_constants():
+    input = r"""<?
+    class A { public $b = self::C; function d($var1=self::C) {} }
+    ?>"""
+    expected = [
+        Class('A', None, None, [], [],
+            [ClassVariables(['public'], [ClassVariable('$b', StaticProperty('self', 'C'))]),
+             Method('d', [], [FormalParameter('$var1', StaticProperty('self', 'C'), False, None)], [], False)
+            ])]
+    eq_ast(input, expected)
+
+def test_backtick_shell_exec():
+    input = '<? `$cmd` . `date`; `echo $line`; ?>'
+    expected = [
+        BinaryOp('.',
+            FunctionCall('shell_exec', [Parameter(Variable('$cmd'), False)]),
+            FunctionCall('shell_exec', [Parameter('date', False)])
+        ),
+        FunctionCall('shell_exec', [Parameter(BinaryOp('.', 'echo ', Variable('$line')), False)])
+    ]
+    eq_ast(input, expected)
+
+def test_open_close_tags_ignore():
+    # The filtered lexer should correctly interpret ?><?
+    input = '<? if (1): if (2) 3; ?><? else: 0; endif;'
+    expected = [
+        If(1, Block([If(2, 3, [], None), None]), [], Else(Block([0])))
+    ]
+    eq_ast(input, expected)
+
+def test_ternary():
+    input = '<? 1 ? 2 : 3; 4 ? : 5;'
+    expected = [
+        TernaryOp(1, 2, 3),
+        TernaryOp(4, 4, 5),
+    ]
+    eq_ast(input, expected)
+
+def test_array_dereferencing():
+    input = '<? $a->method()[0]; func()[1];'
+    expected = [
+        ArrayOffset(MethodCall(Variable('$a'), 'method', []), 0),
+        ArrayOffset(FunctionCall('func', []), 1)
+    ]
+    eq_ast(input, expected)
+
+def test_array_literal():
+    input = '<? [1,2]; [];'
+    expected = [
+        Array([ArrayElement(None, 1, False), ArrayElement(None, 2, False)]),
+        Array([]),
+    ]
+    eq_ast(input, expected)
+
+def test_array_in_default_arg():
+    input = '<? function f($a=[]){} function g($a=array()){}'
+    expected = [
+        Function('f', [FormalParameter('$a', Array([]), False, None)], [], False),
+        Function('g', [FormalParameter('$a', Array([]), False, None)], [], False),
+    ]
+    eq_ast(input, expected)
+
+def test_const_heredoc():
+    input = '''<?
+    const X = <<<HERE
+text
+HERE;'''
+    expected = [
+        ConstantDeclarations([ConstantDeclaration('X', 'text')])
+    ]
+    eq_ast(input, expected)
+
+def test_object_property_on_expr():
+    input = '''<? ($a->m1())->m2(); ($a->m1())->m2;'''
+    expected = [
+        MethodCall(MethodCall(Variable('$a'), 'm1', []), 'm2', []),
+        ObjectProperty(MethodCall(Variable('$a'), 'm1', []), 'm2'),
+    ]
+    eq_ast(input, expected)
+
+def test_binary_string():
+    input = '''<? b"abc"; b'abc';'''
+    expected = [
+        "abc",
+        "abc",
+    ]
+    eq_ast(input, expected)
+
+def test_class_trait_use():
+    input = '''<? class A { use B; }'''
+    expected = [
+        Class('A', None, None, [], [TraitUse('B', [])], []),
+    ]
+    eq_ast(input, expected)
+
+def test_trait():
+    input = '''<? trait A { } trait B { use A; } trait C { function f(){} }
+                  trait D { protected $v; }'''
+    expected = [
+        Trait('A', [], []),
+        Trait('B', [TraitUse('A', [])], []),
+        Trait('C', [], [Method('f', [], [], [], False)]),
+        Trait('D', [], [ClassVariables(['protected'], [ClassVariable('$v', None)])]),
+    ]
+    eq_ast(input, expected)
+
+def test_trait_renames():
+    input = '''<? trait A { use T {X as Y;} }
+                  class B { use T {X as Y;} }
+                  trait C { use T {X as public Y;} }
+                  trait D { use T {X as public;} }
+                  trait E { use T {X::m as Y;} }'''
+    expected = [
+        Trait('A', [TraitUse('T', [TraitModifier('X', 'Y', None)])], []),
+        Class('B', None, None, [], [TraitUse('T', [TraitModifier('X', 'Y', None)])], []),
+        Trait('C', [TraitUse('T', [TraitModifier('X', 'Y', 'public')])], []),
+        Trait('D', [TraitUse('T', [TraitModifier('X', None, 'public')])], []),
+        Trait('E', [TraitUse('T', [TraitModifier(StaticProperty('X', 'm'), 'Y', None)])], []),
+    ]
+    eq_ast(input, expected)
+
+def test_class_name_as_string():
+    input = '''<? A::class; const C = A::class;'''
+    expected = [
+        'A',
+        ConstantDeclarations([ConstantDeclaration('C', 'A')]),
+    ]
+    eq_ast(input, expected)
+
+def test_static_expressions():
+    input = '''<? const C = 1+2; const C = 1+(2+3); const C = "a"."b";'''
+    expected = [
+        ConstantDeclarations([ConstantDeclaration('C', BinaryOp('+', 1, 2))]),
+        ConstantDeclarations([ConstantDeclaration('C', BinaryOp('+', 1, BinaryOp('+', 2, 3)))]),
+        ConstantDeclarations([ConstantDeclaration('C', BinaryOp('.', 'a', 'b'))]),
+    ]
+    eq_ast(input, expected)
+
+def test_const_arrays():
+    input = '''<? const C = array(1+2);'''
+    expected = [
+        ConstantDeclarations([ConstantDeclaration('C', Array([ArrayElement(None, BinaryOp('+', 1, 2), False)]))]),
+    ]
+    eq_ast(input, expected)
+
+def test_numbers():
+    input = '''<? 10; 010; 0x10; 0b10;'''
+    expected = [
+        10,
+        0o10,
+        0x10,
+        2,
+    ]
+    eq_ast(input, expected)
+
+def test_result_multiple_offsets():
+    input = '''<? $o->m()[1][2]; $o->m(){1}{2}; '''
+    expected = [
+        ArrayOffset(ArrayOffset(MethodCall(Variable('$o'), 'm', []), 1), 2),
+        StringOffset(StringOffset(MethodCall(Variable('$o'), 'm', []), 1), 2),
+    ]
+    eq_ast(input, expected)
+
+def test_yield():
+    input = '''<? function f() { yield; yield 1; }'''
+    expected = [
+        Function('f', [], [
+            Yield(None),
+            Yield(1),
+        ], False),
+    ]
+    eq_ast(input, expected)
+
+def test_static_property_dynamic_access():
+    input = '''<? $o::{$prop};'''
+    expected = [
+        StaticProperty(Variable('$o'), Variable('$prop')),
+    ]
+    eq_ast(input, expected)
+
+def test_static_property_dynamic_call():
+    input = '''<? $o::{$prop}();'''
+    expected = [
+        StaticMethodCall(Variable('$o'), Variable('$prop'), []),
+    ]
+    eq_ast(input, expected)
+
+def test_nowdoc():
+    input = r"""<?
+        echo <<<'EOT'
+disregard $all {$crazy} ${stuff}->f();
+and `this`
+EOT;
+    ?>"""
+    expected = [
+        Echo(['disregard $all {$crazy} ${stuff}->f();\nand `this`'])
+    ]
+    eq_ast(input, expected)
+
+def test_exit_loc():
+    input = '''<?
+               exit(1); '''
+    expected = [
+        Exit(1, 'exit', lineno=2)
+    ]
+    eq_ast(input, expected, with_top_lineno=True)
